from abc import ABC, abstractmethod
from typing import List

from .base_models import Context, Concept
from .in_close import inclose_start
from .association_rules import get_association_rules
from .border_hasse import calculate_hasse


class FCASolver(ABC):
    
    @abstractmethod
    def get_concepts(self, ctx : Context) -> List[Concept]:
        return None

    def get_lattice(self, ctx : Context) -> (List[List[int]], List[Concept]):
        """Returns a graph with indexes and a list of the concepts by id.
        """
        concepts = self.get_concepts(ctx)
        return calculate_hasse(ctx, concepts)

    def get_association_rules(self, ctx, min_support=0.5, min_confidence=1):
        return get_association_rules(self.get_concepts(ctx), min_support, min_confidence)


class Inclose(FCASolver):

    def get_concepts(self, ctx : Context) -> List[Concept]:
        res = inclose_start(ctx) # one of the caveats of inclose is that it doesn't include bottom = (A', A)
        if not self._has_bottom(ctx, res):
            res.append(self._calculate_bottom(ctx))
        return res
    

    def _calculate_bottom(self, ctx : Context) -> Concept:
        objects_with_all_attributes = []
        for i in range(len(ctx.O)):
            is_related_to_all = True
            for j in range(len(ctx.A)):
                if not ctx.I[i][j]:
                    is_related_to_all = False
                    break
            if is_related_to_all:
                objects_with_all_attributes.append(i)
        return Concept(ctx, objects_with_all_attributes, [j for j in range(len(ctx.A))])
    
    def _has_bottom(self, ctx, concepts) -> bool:
        for concept in concepts:
            if len(concept.A) == ctx.A:
                return True
        return False





#!/usr/bin/env python

import os
import argparse
import csv


INVALID_FILETYPE_MSG = "Error: Invalid file format. {} must be a .txt file."
INVALID_PATH_MSG = "Error: Invalid file path/name. Path {} does not exist."
  
  
def validate_file(file_name):
    '''
    validate file name and path.
    '''
    if not valid_path(file_name):
        print(INVALID_PATH_MSG.format(file_name))
        quit()
    elif not valid_filetype(file_name):
        print(INVALID_FILETYPE_MSG.format(file_name))
        quit()
    return
      
def valid_filetype(file_name):
    # validate file type
    return file_name.endswith('.csv')
  
def valid_path(path):
    # validate file path
    return os.path.exists(path)


def main():
    # create a parser object
    parser = argparse.ArgumentParser(description = "FCA cli")
    
    # add argument
    parser.add_argument("--rca", type=bool, nargs='?',
                        metavar="apply_rca",
                        help="True if the RCA is to be used", default=False)
    
    parser.add_argument("-c", "--context", type=str, nargs=1,
                        metavar="context_name",
                        help="Formal context csv file.")
    
    parser.add_argument("-k", "--contexts", type=str, nargs='*',
                        metavar="context_names",
                        help="Formal contexts csv files from the relational context family.")
    
    parser.add_argument("-r", "--relations", type=str, nargs='*',
                        metavar="relation_file_names",
                        help="Relation csv filename in case of RCA. Name is expected to be r_0_3.csv for example if its a "
                             "relation between objects of the contexts 0 and 3 respectively")
    
    # parse the arguments from standard input
    args = parser.parse_args()

    if args.rca:
        K, R = parse_rca(args)
        print(f'### RCA parsed {K}, {R}')
    else:
        K = parse_fca(args)
        print(f'### FCA parsed {K}')


def parse_rca(args):
    contexts = []
    for ctx_name in args.contexts:
        contexts.append(import_context(ctx_name))
    
    relations = []
    for relation_name in args.relations:
        relations.append(import_relation(relation_name))
    
    return contexts, relations


def parse_fca(args):
    return import_context(args.context)


def import_context(filename):
    O = []
    A = []
    I = []
    with open(filename, 'r') as f:
        reader = csv.reader(filename)
        loading_attributes = True
        for row in reader:
            if loading_attributes:
                for attr in row[1:]:
                    A.append(attr)
                loading_attributes = False
            else:
                O.append(row[0])
                I.append([])
                for attr_i, i in enumerate(row[1:]):
                    I.append(len(attr_i) != 0)
    return O, A, I


def import_relation(filename, contexts):
    contexts_indexes = [int(idx) for idx in filename.split('_')[1:]]
    R = dimension(contexts, contexts_indexes, 0)
    with open(filename, 'r') as f:
        reader = csv.reader(filename)
        for row in reader:
            to_process = len(row) - 1
            current = R[0]
            while to_process > 0:
                i = len(row) - to_process
                current = current[int(row[i])]
                to_process -= 1
            current.add(int(row[-1]))  # this should be a set
    return R
            

def dimension(contexts, contexts_indexes, i):
    if i == len(contexts_indexes[i].O) - 1:
        return set()
    return [dimension(contexts, contexts_indexes, i+1) for _ in range(len(contexts_indexes[i].O))]


if __name__ == "__main__":
    main()



from typing import List


class Context:
    def __init__(self, O, A, I):
        """
        :param O: a list of objects (strings for the moment) len(O) = n
        :param A: a list of attributes (strings for the moment) len(A) = m
        :param I: a nm incidence matrix, beign I[i][j] = 1 if object i has the attribute j, 0 <= i <= n, 0 <= j <= m
        """
        self.O : List = O
        self.A : List = A
        self.I : List = I
    
    def derivative(self, Y, is_attr : bool = True):
        """
        :param Y: A subset of A
        """
        return self._derivative_attr(Y) if is_attr else self._derivative_obj(Y)

    def _derivative_attr(self, Y):
        res = set()
        for attr in Y:
            attr_idx = self.A.index(attr)
            for obj in self.O:
                obj_idx = self.O.index(obj) 
                if I[obj_idx][attr_idx]:
                    res.add(obj)
        return list(res)
    
    def _derivative_obj(self, Objs):
        res = set()
        for obj in Objs:
            obj_idx = self.O.index(obj) 
            for attr in self.A:
                attr_idx = self.A.index(attr)
                if I[obj_idx][attr_idx]:
                    res.add(attr)
        return list(res)

    def __repr__(self):
        return f'O: {self.O},\n' \
               f'A: {self.A},\n' \
               f'I: {self.I}'


class Concept:
    def __init__(self, context : Context, O, A, parents=None, children=None):
        if children is None:
            children = []
        
        if parents is None:
            parents = []

        self.context = context
        self.O = O
        self.A = A
        self._set_O = frozenset(O)  # this is not supposed to change, so it's a frozenset
        self.parents = parents
        self.children = children
    
    def in_extent(o) -> bool:
        return o in self._set_O  # O(1) amortised
    
    def add_child(self, concept):
        self.children.append(concept)
        concept.parents.append(self)
    
    def __repr__(self):
        return f'({[self.context.O[i] for i in self.O]}, {[self.context.A[i] for i in self.A]})'
    
    def to_tuple(self):
        return [self.context.O[i] for i in self.O], [self.context.A[i] for i in self.A]




def calculate_hasse(ctx, concepts):
    """Calculates the hasse graph and outputs it as:
      List[List[int]], ordered_concepts
      Note that the indexes in the hasse are the indexes in the ordered_concepts.
    """
    if not concepts:
        return []

    number_of_attributes = len(ctx.A)

    ordered_concepts = order(concepts, number_of_attributes)  # creates a new list with the concepts ordered based on the border algorithm
    hasse_lattice = [[] for _ in range(len(concepts))]  # a graph
    border = set([0])
    for i in range(1, len(concepts)):
        ci = ordered_concepts[i]
        intent_ci = frozenset(ci.A)
        intents = []
        concepts_by_intent = {}
        for c_idx in border:
            c = ordered_concepts[c_idx]
            intents_intersected = set(c.A)
            intents_intersected = frozenset(intents_intersected.intersection(intent_ci))
            intents.append(intents_intersected)
            concepts_by_intent[intents_intersected] = c_idx
        cover_intents = maxima(intents, number_of_attributes)
        upper_covers = set()
        for intent in cover_intents:
            idx = find_concept(intent, concepts_by_intent, ordered_concepts, i)
            c = ordered_concepts[idx]
            ci.add_child(c)
            hasse_lattice[i].append(idx)
            upper_covers.add(idx)
        border.difference_update(upper_covers)
        border.add(i)
    return hasse_lattice, ordered_concepts
            



def order(concepts, number_of_attributes):
    """linear time sorting (in |concpets|) of concepts
    """
    bunches = [[] for _ in range(number_of_attributes + 1)]
    res = []
    for c in concepts:
        bunches[len(c.A)].append(c)
    return [concept for bunch in bunches for concept in bunch]


def maxima(intents, number_of_attributes):
    ordered_intents = order_intents(intents, number_of_attributes)
    maxi = []
    for intent in ordered_intents:
        is_min = True
        for elem in maxi:
            is_min = is_min and not intent.issubset(elem)
        if is_min:
            maxi.append(intent)
    return maxi


def order_intents(intents, number_of_attributes):
    bunches = [[] for _ in range(number_of_attributes + 1)]
    res = []
    for intent in intents:
        bunches[len(intent)].append(intent)
    
    return [intent for bunch in reversed(bunches) for intent in bunch]


def find_concept(intent, cover_intents, ordered_concepts, idx_so_far):
    i = cover_intents[intent]
    c = ordered_concepts[cover_intents[intent]]
    Y = set(intent)
    while set(c.A) != Y:
        i -= 1
        c_aux = ordered_concepts[i]  # at least it'll have one
        while not Y.issubset(set(c_aux.A)):
            i -= 1
            c_aux = ordered_concepts[i]
        c = c_aux
    return i
        


import pytest
from src.fca.rca.p18n import EXISTS
from src.fca.rca.rca import rca_get_relations, lattice_and_concept_idx
from src.fca.api_models import Context
from src.fca.plot.plot import plot_from_hasse


def test_rca_converges(create_objects_and_attributes, create_objects_and_attributes_2, create_relation_between_0_and_1):
    k_1 = create_objects_and_attributes
    k_2 = create_objects_and_attributes_2
    K = [Context(*k_1), Context(*k_2)]
    R = [create_relation_between_0_and_1]
    lattices = rca_get_relations(K, R, EXISTS)
    assert len(lattices) == 2


def test_rca_exists_makes_sense(create_objects_and_attributes, create_objects_and_attributes_2, create_relation_between_0_and_1):
    relation = create_relation_between_0_and_1
    k_1 = Context(*create_objects_and_attributes)
    number_of_k_1_0_attibutes = len(k_1.A)
    k_2 = Context(*create_objects_and_attributes_2)
    K = [k_1, k_2]
    R = [create_relation_between_0_and_1]
    lattices = rca_get_relations(K, R, EXISTS)
    lattice_k_1 = lattices[0]
    lattice_k_2 = lattices[1]
    exists_symbol = str(EXISTS)
    relation_symbol_idx = number_of_k_1_0_attibutes
    while relation_symbol_idx < len(k_1.A):
        relational_attribute = k_1.A[relation_symbol_idx]
        i, j = lattice_and_concept_idx(relational_attribute)
        for obj_idx, o in enumerate(k_1.O):
            found_one = exists_relation(lattice_k_2, j, relation, obj_idx)
            assert (k_1.I[obj_idx][relation_symbol_idx] and found_one) or \
                   (not k_1.I[obj_idx][relation_symbol_idx] and not found_one)
        relation_symbol_idx += 1


def test_rca_plot(create_objects_and_attributes, create_objects_and_attributes_2, create_relation_between_0_and_1):
    k_1 = create_objects_and_attributes
    k_2 = create_objects_and_attributes_2
    K = [Context(*k_1), Context(*k_2)]
    R = [create_relation_between_0_and_1]
    lattices = rca_get_relations(K, R, EXISTS)
    for lattice in lattices:
        hasse, concepts = lattice.hasse, lattice.concepts
        plot_from_hasse(hasse, concepts)
    assert lattices


def exists_relation(lattice_k_2, j, relation, obj_idx):
    found_one = False
    for o_2 in lattice_k_2.concepts[j].O:
        found_one = found_one or o_2 in relation[obj_idx]
    return found_one



begin_tikz = """
\\documentclass[tikz,border=2mm]{standalone}

\\tikzset{white border/.style={preaction={draw,white,line width=4pt}}}

\\newcommand{\\nt}[1]{$\langle#1\\rangle$} % node text

\\begin{document}
\\begin{tikzpicture}[x=1.4cm,y=1.8cm]% <-- change this numbers if you need (separations)
% nodes
"""

end_tikz = """
\\end{tikzpicture}
\\end{document}
"""


from ..utils.utils import from_subscript


def rca_get_relations(K, R, p):
    """
    :param K: contexts
    :param R: relations
    :param p: p18n operator
    :return: Relational Lattice family
    """
    lattices = []
    calculating_lattices = []
    for k in K:
        lattices.append(k.get_lattice())
        calculating_lattices.append(None)
    
    stable = False
    while not stable:
        # Invariant:
        #   lattices_prev has the last lattices calculated
        for i, k in enumerate(K):
            for r in R:
                if i == r.i:
                    k.graduate(r, p, lattices[r.j], r.j)
        
        for i, k in enumerate(K):
            calculating_lattices[i] = k.get_lattice()
        
        # O(|calculating_lattices| * |concepts|)
        for i, l in enumerate(calculating_lattices):
            stable = stable or l.isomorph(lattices[i])  # O(|concepts|)
            lattices[i] = l  # maintains the invariant
    return lattices


def lattice_and_concept_idx(relational_attribute):
    double_points_idx = relational_attribute.index('C')
    without_C = relational_attribute[double_points_idx + 1:]
    i, j = without_C.split('₋')
    return from_subscript(i), from_subscript(j)



def rca_get_relations(K, R, P):
    """
    :param K: relational context family
    :param R: relations
    :param P: array with p18n operators for each R (len(P) == len(R))
    :return: Relational Lattice family
    """
    prev_concepts_len = 0
    curr_concepts_len = None
    lattices = []
    calculating_lattices = []
    for k in K:
        lattices.append(k.get_lattice())
        prev_concepts_len += lattices[-1].concepts_len()
        calculating_lattices.append(None)
    
    while prev_concepts_len != curr_concepts_len:
        # Invariant:
        #   lattices_prev has the last lattices calculated
        prev_concepts_len = curr_concepts_len or prev_concepts_len
        for k in K:
            for j, r in enumerate(R):
                for p in P[j]:
                    k.graduate(r, p, lattices[j])
        
        for i, k in enumerate(K):
            calculating_lattices[i] = k.get_lattice()
        
        # maintains the invariant, compress, and canonize
        curr_concepts_len = 0
        for i, l in calculating_lattices:
            lattices[i] = l.compress().canonize()
            curr_concepts_len += lattices[i].concepts_len()

    return lattices



import pytest
from src.fca.base_models import Context
from src.fca.in_close import inclose_start
from src.fca.get_lattice import Inclose
from src.fca.plot.plot import plot_from_hasse



def test_inclose_returns_all_the_concepts_but_bottom(create_objects_and_attributes):
    O, A, I = create_objects_and_attributes
    c = Context(O, A, I)
    concepts = inclose_start(c)
    res = [x.to_tuple() for x in concepts]
    expected_result = [(['Tg05', 'Tg05FX', 'Flxtra', 'Hxr'], []),
                       (['Tg05FX', 'Hxr'], ['Effect', 'Control']),
                       (['Tg05FX'], ['Effect', 'Control', 'Short']),
                       (['Tg05'], ['Power', 'Short']),
                       (['Tg05FX', 'Flxtra', 'Hxr'], ['Control']),
                       (['Tg05', 'Tg05FX'], ['Short'])]
    assert res == expected_result


def test_inclose_solver_returns_all_the_concepts(create_objects_and_attributes):
    O, A, I = create_objects_and_attributes
    c = Context(O, A, I)
    concepts = Inclose().get_concepts(c)
    res = [x.to_tuple() for x in concepts]
    expected_result = [(['Tg05', 'Tg05FX', 'Flxtra', 'Hxr'], []),
                       (['Tg05FX', 'Hxr'], ['Effect', 'Control']),
                       (['Tg05FX'], ['Effect', 'Control', 'Short']),
                       (['Tg05'], ['Power', 'Short']),
                       (['Tg05FX', 'Flxtra', 'Hxr'], ['Control']),
                       (['Tg05', 'Tg05FX'], ['Short']),
                       ([], ['Effect', 'Power', 'Control', 'Short'])]
    assert res == expected_result


def test_inclose_solver_returns_all_the_concepts_when_there_are_objects_with_all_attributes(create_objects_and_attributes):
    O, A, I = create_objects_and_attributes
    c = Context(O, A, I)
    c.I[1][1] = True
    concepts = Inclose().get_concepts(c)
    res = [x.to_tuple() for x in concepts]
    assert (['Tg05FX'], ['Effect', 'Power', 'Control', 'Short']) in res


def test_inclose_solver_returns_association_rules(create_objects_and_attributes):
    O, A, I = create_objects_and_attributes
    c = Context(O, A, I)
    assoc_rules = [r for r in Inclose().get_association_rules(c, min_support=0.4, min_confidence=1)]
    assert len(assoc_rules) == 1
    first_association_rules = assoc_rules[0]
    x = first_association_rules.ordered_statistics[0]

    assert 0.42 < first_association_rules.support < 0.43

    # Effect -> Control
    assert len(x.items_base) == 1
    assert 0 in x.items_base

    assert len(x.items_add) == 1
    assert 2 in x.items_add

    assert x.confidence == 1


def test_lattice_construction(create_objects_and_attributes_2):
    O, A, I = create_objects_and_attributes_2
    c = Context(O, A, I)
    hasse_lattice, concepts = Inclose().get_lattice(c)
    plot_from_hasse(hasse_lattice, concepts)
    assert hasse_lattice



import networkx as nx
from collections import deque

from ..border_hasse import calculate_hasse
from matplotlib import pyplot as plt


def plot_from_hasse(hasse, concepts_by_id, show_plot=True, save_plot : str=None):
    G = nx.DiGraph()
    for u in range(len(hasse)):
        G.add_node(to_node(u, concepts_by_id))

    for u in range(len(hasse)):
        for w in hasse[u]:
            G.add_edge(to_node(u, concepts_by_id), to_node(w, concepts_by_id))
    
    pos = dict()
    distances, diameter = _compute_distances(hasse, from_node=len(hasse) - 1)
    middle = diameter / 2
    for i in range(len(distances)):
        level_i_len = len(distances[i])
        new_middle = level_i_len / 2
        for idx, node in enumerate(distances[i]):
            horizontal_position = idx / level_i_len
            pos[to_node(node, concepts_by_id)] = (horizontal_position * level_i_len - new_middle, i)
    
    pos[to_node(0, concepts_by_id)] = (pos[to_node(len(hasse) - 1, concepts_by_id)][0],
                                       len(distances) + (1 if len(distances[-1]) == 0 else 0))

    nx.draw_networkx_edges(G, pos, edge_color='blue', arrowsize=13, width=1.3)
    nx.draw_networkx_labels(G, pos, font_weight='bold')
    # plt.savefig('this.png')
    if show_plot:
        plt.show()

    if save_plot:
        nx.nx_pydot.write_dot(G, save_plot if save_plot.endswith(".dot") else f"{save_plot}.dot")


def _compute_distances(hasse, from_node=0):
    distances = [None for _ in range(len(hasse))]
    distances[from_node] = 0
    queue = deque([from_node])
    maximum_distance = 0
    while queue:
        u = queue.popleft()
        for w in hasse[u]:
            if distances[w] is None:
                distances[w] = distances[u] + 1
                maximum_distance = max(maximum_distance, distances[w])
                queue.append(w)
    res_distances = [[] for _ in range(maximum_distance + 1)]
    diameter = 0
    for node in range(len(hasse)):
        res_distances[distances[node]].append(node)
        diameter = max(diameter, len(res_distances[distances[node]]))
    return res_distances, diameter

def to_node(u, concepts_by_id):
    return str(concepts_by_id[u])



from typing import List


from apyori import apriori
from .base_models import Concept

def get_association_rules(concepts : List[Concept], min_support=0.5, min_confidence=1):
    """Given a list of concepts, it returns a generator of its association rules
    """
    transactions = [c.A for c in concepts]
    return apriori(transactions, min_support=min_support, min_confidence=min_confidence)



from .base_models import Context, Concept
from .utils.utils import inverse_range, is_in, lower_bound, insert_ordered


r_new = 0
concepts = []
graph = []


def inclose_start(context : Context):
    global r_new
    global concepts
    global ctx
    ctx = context
    initial_concept = Concept(ctx, [i for i in range(len(context.O))], [])  # Concepts only uses indexes
    concepts = [initial_concept]
    r_new = 0
    inclose(0, 0)

    # pop the last concept if it hasn't been closed
    if len(concepts[-1].A) == 0:
        concepts.pop()

    return concepts


def inclose(r, y):
    """incrementally close the concept r, beginning with attribute y
    """
    global r_new
    global concepts
    global ctx
    r_new += 1
    for j in range(y, len(ctx.A)):
        if len(concepts) <= r_new:
            concepts.append(Concept(ctx, [], []))
        else:
            concepts[r_new].O = []

        for i in concepts[r].O:  # concepts[r].O \cap {j}'
            if ctx.I[i][j]:
                is_there, _ = is_in(i, concepts[r_new].O)
                if not is_there:
                    insert_ordered(i, concepts[r_new].O)  # append only if it's not in the list
        
        if len(concepts[r_new].O) > 0:
            if len(concepts[r_new].O) == len(concepts[r].O):
                is_there, _ = is_in(j, concepts[r].A)
                if not is_there:
                    insert_ordered(j, concepts[r].A)
            elif is_canonical(r, j - 1):
                new_attributes = concepts[r].A.copy()
                is_there, _ = is_in(j, new_attributes)
                if not is_there:
                    insert_ordered(j, new_attributes)
                concepts[r_new].A = new_attributes  # O(|concepts[r].A|)
                inclose(r_new, j + 1)


def is_canonical(r, y):
    new_y = y
    for k in inverse_range(len(concepts[r].A)):
        for j in range(new_y, concepts[r].A[k], -1):
            h = 0
            while h < len(concepts[r_new].O):
                if not ctx.I[concepts[r_new].O[h]][j]:
                    break
                h += 1
            if h == len(concepts[r_new].O):
                return False
        new_y = concepts[r].A[k] - 1
    
    for j in range(new_y, -1, -1):
        h = 0
        while h < len(concepts[r_new].O):
            if not ctx.I[concepts[r_new].O[h]][j]:
                break
            h += 1
        if h == len(concepts[r_new].O):
            return False
    return True



"""
Propositionalization (P18N) operators
"""

from ..utils.utils import to_subscript


class Propositionalization:
    def __call__(self, object, relation, concept) -> bool:
        return False


class Exists(Propositionalization):
    def __call__(self, obj, relation, concept) -> bool:
        for o in concept.O:
            # with a set structure or bitset, we would know this in
            # O(len(objects)) / 64, considering a 64bit arc
            # other option would be to use a set and then relation[i][j] returns whether j is in the set relation[i],
            # but this latter option is costly in memory, although easier to implement.
            if o in relation[obj]:
                return True
        return False
    
    def __repr__(self):
        return '∃'


class Forall(Propositionalization):
    def __call__(self, obj, relation, concept) -> bool:
        for o in concept.O:
            # with a set structure or bitset, we would know this in
            # O(len(objects)) / 64, considering a 64bit arc
            # other option would be to use a set and then relation[i][j] returns whether j is in the set relation[i],
            # but this latter option is costly in memory, although easier to implement.
            if o not in relation[obj]:
                return False
        return True
    
    def __repr__(self):
        return '∀'


class Relation(list):
    def __init__(self, array, i, j):
        super().__init__(array)
        self.i = i
        self.j = j

    def __repr__(self):
        return f'R{to_subscript(self.i)}₋{to_subscript(self.j)}'




from .base_models import Context as Ctx
from .get_lattice import FCASolver, Inclose
from .utils.utils import to_subscript


class Context(Ctx):
    def __init__(self, O, A, I, solver : FCASolver = None):
        super().__init__(O, A, I)
        
        if solver is None:
            solver = Inclose()
        
        self.solver = solver
        self.iteration = 0

        # FIXME: The ideal would be to cache the last_lattice
        #        and also calculate them from the point they were already calculated
        #        * is that possible not to repeat calculations?
        #        * if not, can we bound the amount of calculations that have to be repeated?
        self._last_lattice = None
    
    def get_concepts(self):
        return self.solver.get_concepts(self)
    
    def get_lattice(self):
        hasse, concepts = self.solver.get_lattice(self)
        if self._last_lattice is None:
            self._last_lattice = Lattice(hasse, concepts)
        else:
            from_idx = len(self._last_lattice.concepts)
            self._last_lattice.from_iteration.append(from_idx)
            self._last_lattice = Lattice(hasse, concepts, self._last_lattice.from_iteration)
        return self._last_lattice
    
    def graduate(self, relation, p, lattice, lattice_idx):
        """
        Applies graduation. Extends the formal context with more attributes
        with the relation `relation`, using the p18n operator, against the lattice `lattice`
        """
        q = self.iteration
        for i, concept in lattice.concepts_from(q):
            self.A.append(f'{p}{relation} : C{to_subscript(lattice_idx)}₋{to_subscript(i)}')
        
        for o, relations in enumerate(self.I):
            # this only works considering the relations is a array[array[bool]], if we wanted to make this stream-friendly,
            # we'd have to change it perhaps.
            relations.extend([p(o, relation, c) for _, c in lattice.concepts_from(q)])
        
        self.iteration += 1


class Lattice:
    def __init__(self, hasse, concepts, from_iteration=None):
        self.hasse = hasse
        self.concepts = concepts
        self.ctx = self.concepts[-1].context
        if from_iteration is None:
            # from_iteration[q] gives the idx in `concepts` from where to start iterating the concepts
            # added in iteration q
            self.from_iteration = [0]
        else:
            self.from_iteration = from_iteration
    
    def isomorph(self, other_lattice):
        """
        This method needs the two lattices to have the concepts ordered in the same way, othetwise it'll fail
        - Complexity: O(|self.concepts|), \omega(1)
        """
        if len(self.concepts) != len(other_lattice.concepts):
            return False
        
        for i, concept in enumerate(self.concepts):
            if len(concept.O) != len(other_lattice.concepts[i].O) or \
               len(concept.A) != len(other_lattice.concepts[i].A):
                return False
        return True
    
    def concepts_from(self, q):
        """iterator for all the concepts of iteration q and their indexes: idx, concept
        """
        from_i = self.from_iteration[q]
        to_i = len(self.concepts) if len(self.from_iteration) <= q + 1 else self.from_iteration[q + 1]
        for i in range(from_i, to_i):
            yield i, self.concepts[i]
    
    def __repr__(self):
        return str(self.concepts)








import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fca-algorithms",
    version="0.0.14",
    author="Ramshell",
    author_email="ramshellcinox@gmail.com",
    license="CC By 4.0",
    description="FCA basic algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    entry_points={
        'console_scripts': [
            'fca_cli=fca.scripts.fca_cli:main'
        ]
    },
)


"""Closed partially-ordered patterns impl
"""
from collections import deque
from .rca import lattice_and_concept_idx


def get_index_offsets(lattices):
    lattice_idx_offsets = [0]  # lattice 0 has 0 offset
    for idx in len(len(lattices) - 1):
        lattice_idx_offsets.append(lattice_idx_offsets[-1] + len(lattices[idx].concepts))
    return lattice_idx_offsets


def create_cpop(lattices):
    """Returns a graph with the cpop.
    Note that indexes in the graph will always be `idx - previous_lattice_last_index`
    """
    lattice_idx_offsets = get_index_offsets(lattices)
    graph = create_graph(lattices, lattice_idx_offsets)
    # at this point, graph contains all the lattices unconected between them
    create_relational_edges(graph, lattices, lattice_idx_offsets)
    # here the graph is completed in the sense that its relational edges has already been added
    return graph


def get_subgraph_from_concept(graph, concept_idx) -> set:
    nodes_in_connected_component = set([concept_idx])
    visited = set()
    q = deque([concept_idx])
    while q:
        u = q.popleft()
        visited.add(u)
        for v, is_relational in graph[u]:
            if v not in visited and is_relational:
                nodes_in_connected_component.add(v)
                q.append(v)
    return nodes_in_connected_component


def create_graph(lattices, idx_offsets):
    graph_so_far = []
    for i, lattice in enumerate(lattices):
        for neighbours in lattice.hasse:
            # Edges are (node, bool) where bool is False if the edge is not relational and True otherwise
            # offset all neighbours
            graph_so_far.append([(idx_offsets[i] + neighbour, False) for neighbour in neighbours])
    return graph_so_far


def create_relational_edges(graph, lattices, lattice_idx_offsets):
    for i, lattice in enumerate(lattices):
        for c_idx, concept in enumerate(lattice.concepts):
            idx_in_graph = lattice_idx_offsets[i] + c_idx
            for attr in concept.A:
                if is_attr_relational(attr):
                    i_related, j = lattice_and_concept_idx(attr)
                    concept_idx_related = lattice_idx_offsets[i_related] + j
                    graph[concept_idx_related].append((concept_idx_related, True))  # These are always relational


def is_attr_relational(attr):
    return ':' in attr  # this can be highly improved


def is_edge_relational(edge):
    return edge[1]



import pytest

from src.fca.rca.models import Relation

@pytest.fixture
def create_objects_and_attributes():
    O = ['Tg05', 'Tg05FX', 'Flxtra', 'Hxr']
    A = ['Effect', 'Power', 'Control', 'Short']
    I = [
        [False, True, False, True],
        [True, False, True, True],
        [False, False, True, False],
        [True, False, True, False],
    ]
    return O, A, I


@pytest.fixture
def create_objects_and_attributes_2():
    O = ['1', '2', '3', '4', '5']
    A = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i']
    I = [
        [True, False, True, False, False, True, False, True, False],
        [True, False, True, False, False, False, True, False, True],
        [True, False, False, True, False, False, True, False, True],
        [False, True, True, False, False, True, False, True, False],
        [False, True, False, False, True, False, True, False, False],
    ]
    return O, A, I


@pytest.fixture
def create_relation_between_0_and_1():
    I = [
        set([0, 2]),  # object in index 0 is related with 0 and also with 2 of the other context
        set([0, 2, 4]),  # object in index 1 is related with 0, 2 and 4 of the other context
        set([0, 3]),
        set([1, 2]),
    ]
    return Relation(I, 0, 1)



def inverse_range(length, stop=-1):
    """helper function to iterate in an inverse order
    """
    return range(length - 1, stop, -1)


def is_in(elem, array):
    """given an `elem` and an ordered array, it returns True and its index if the element is in there,
    False, None otherwise
    """
    low = 0
    high = len(array)
    while low < high:
        mid = (low + high) // 2
        if array[mid] < elem:
            low = mid + 1
        else:
            high = mid

    if low < len(array) and low >= 0 and array[low] == elem:
        return True, low
    return False, None


def lower_bound(elem, array):
    """returns the index in which you should insert the element `elem` to keep the order
    """
    low = 0
    high = len(array)
    while low < high:
        mid = (low + high) // 2
        if array[mid] < elem:
            low = mid + 1
        else:
            high = mid

    return low


def insert_ordered(elem, array):
    """inserts `elem` in `array` in an index that keeps the order
    """
    array.insert(lower_bound(elem, array), elem)


subscripts = '₀₁₂₃₄₅₆₇₈₉'  # outside the function so that it mallocs only once (?)
def to_subscript(number):
    if number < 0:
        return ValueError(f"{number} should be positive or zero")

    res = [x for x in str(number)]
    i = len(res) - 1
    while i >= 0:
        res[i] = subscripts[int(res[i])]
        i -= 1
    return ''.join(res)


integer_numbers = {
    '₀': '0',
    '₁': '1',
    '₂': '2',
    '₃': '3',
    '₄': '4',
    '₅': '5',
    '₆': '6',
    '₇': '7',
    '₈': '8',
    '₉': '9',

}  # outside the function so that it mallocs only once (?)
def from_subscript(number):
    res = [x for x in number]
    i = len(res) - 1
    while i >= 0:
        res[i] = integer_numbers[res[i]]
        i -= 1
    return int(''.join(res))
